--------------------------------------------------------------------------------
--
-- Name:         EuDoc - Euphoria documentation program
-- Version:      0.1
--
-- Author:       D. Newhall
--
-- License:      Mozilla
--
-- Created:      Feb-7-2005
-- Updated:      Feb-14-2005
-- Updated:      Jul-13-2005
-- Updated:      Jul-29-2005 Separated EuDocs.ex to EuDoc.ex and EuLib.e
--
-- Intro:
--  EuDoc is a program for automatically creating documentation
--  for Euphoria libraries. The file created has the same visual
--  appearance as the library routines list in the official Euphoria
--  documentation. It can also be used to document programs but in a
--  more limited scope.    <br> <br>
--
--  For EuDoc to work you first need to document your program using these tags.
--  The tags are case insensitive and need not be in any order except that 
--  "Name&#58;" and "Routine&#58;" must come before a block of info relating to
--  that but it's highly
--  recommended that you use the capitalization scheme and order shown for
--  better clarity. Also, to show emphasis or to do special formatting you just
--  need to put the desired HTML tags in your comments.
--
--  <pre>
--
--  File descriptor tags:
--
--    Name&#58;            Name of library/program. Required first
--    Author&#58;          Author's name
--    Created&#58;         Date file was created
--    Updated&#58;         Date file was updated
--    License&#58;         License library's source code is released under
--    Intro&#58;           Introduction comments about the library
--
--  Routine descriptor tags:
--
--    Routine&#58;         Name of the routine. Required first
--    Author&#58;          Author of the routine
--    Created&#58;         Date routine was created
--    Updated&#58;         Date routine was updated and contributor who did so
--    Platform&#58;        Platform the routine runs on
--                         (General, WIN32, DOS32, Linux, FreeBSD)
--    Syntax&#58;          An illustration of the usage syntax
--    Description&#58;     A description of what the routine does
--    Comments&#58;        Extra comments about the routine
--    Example&#58;         A more example to illustrate usage
--    Example Program&#58; The name/location of an example program if such
--                          a program exists
--    See Also&#58;        Lists other related routines in the library
--
--  </pre>
--
--  To try out EuDoc type "ex EuDoc.ex EuDocLib.e" (remember to substitute "ex"
--  for the name of the interpreter on your platform: "exu" for Linux and
--  FreeBSD and "exw" for Windows). EuDocLib.e is the file that contains all
--  the routines for use in EuDoc.ex. It's commented so as to produce the
--  documentation you are reading right now! Besides the information on running
--  EuDoc this file also contains documentation for the routines used by
--  EuDoc as defined in EuDocLib.e as an example.  <br>  <br>
-- 
--  Please send all your comments, questions, suggestions, and bug reports to
--  <a href="mailto: derek_newhall@yahoo.com">Derek Newhall</a>.
--
--------------------------------------------------------------------------------


-- Include files

include wildcard.e
include get.e
include sort.e




-- Constants

global constant  TRUE = 1,
             FALSE = 0,


             LIBRARY = 1,  -- Denotes that library info is being processed
             ROUTINE = 2,  -- Denotes that routine info is being processed
     
             -- Library info constants (some are shared with routines)
             NAME = 1,
             VERSION = 2,
             AUTHOR = 3,
             CREATED = 4,
             UPDATED = 5,
             LICENSE = 6,
             INTRO = 7,

             -- Routine info constants
             PLATFORM = 5,
             SYNTAX = 6,
             DESCRIPTION = 7,
             COMMENTS = 8,
             EXAMPLE = 9,
             EXAMPLE_PROG = 10,
             SEE_ALSO = 11,

             -- Initial data for routines
             --   (added to Routines when new routine is being parsed)
                 ROUTINE_STUB = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}



-- Global variables

global sequence Library_header,  -- Library header
                Routines,        -- Routine data
                Filenums         -- File handles for files

global integer Parsing,     -- TRUE if program is parsing data, FALSE otherwise
               Parse_type,  -- Type of data to aquire (see constants)

               Last         -- Last type of information being stored.
                            -- Should be set to the constant representing the
                            -- current field in Routines[...] or Library_header
                            -- where data is being written to



-- Initialize globals

Library_header = repeat("", 7)  -- Initialize Library_header
Routines = {}
Filenums = {}

-- Parsing and Parse_type are initialized later for every file




-------------------------------------------------------------------------------
-- Routine:  error
-- Syntax:   error(s)
-- Description: 
--   Prints out string passed to error and then aborts with an exit status of 1.
--------------------------------------------------------------------------------
global procedure error(sequence msg)

    printf(2, "Error: %s\n", {msg})
    abort(1)
    
end procedure  -- error




--------------------------------------------------------------------------------
-- Routine:  get_command_line_options
-- Syntax:   get_command_line_options()
-- Description:
--   Gets the file names off the command line and sets Filenums to their
--   file handles.
--------------------------------------------------------------------------------
global procedure get_command_line_options()

    sequence cmd
    integer fn

    -- Get arguments off the command line
    cmd = command_line()
    if length(cmd) < 3 then
    puts(2, "Error: Not enough arguments specified.\n\n")
    puts(2, "Usage: ex EuDoc.ex file [file2 ...]\n\n")
    puts(2, "For more documentation run EuDoc on EuDocLib.e.\n\n")
    puts(2, "  ex EuDoc.ex EuDocLib.e\n")
    abort(1)
    elsif length(cmd) > 23 then
    puts(2, "Error: Only 23 files can be documented at once.")
    end if


    -- Takes the file names from the command line and opens them
    for i = 3 to length(cmd) do
    fn = open(cmd[i], "r")
    if fn = -1 then
        printf(2, "Error: file %s doesn't exist.\n", {cmd[i]})
        abort(1)
    else
        Filenums &= fn
    end if
    end for

end procedure  -- get_command_line




--------------------------------------------------------------------------------
-- Routine:  trim
-- Syntax:   s1 = trim(s2)
-- Description:  Removes leading spaces and '-'s from a string.
--------------------------------------------------------------------------------
global function trim(sequence string)

    sequence tmp
    
    tmp = ""

    for i=1 to length(string) do
    if string[i] != ' ' and string[i] != '\t' and string[i] != '-' then
        tmp = string[i..length(string)]
        exit
    -- Else do next iteration
    end if
    end for
    
    -- If line is all white-space return FALSE
    if length(tmp) = 0 then
    return FALSE
    end if
    
    for i=length(tmp) to 1 by -1 do
    if tmp[i] != ' ' and tmp[i] != '\t' and tmp[i] != ':' then
        return tmp[1..i]
    end if
    end for

end function  -- trim


--------------------------------------------------------------------------------
-- Routine:  trim_comment
-- Syntax:   s1 = trim(s2)
-- Description:
--   Same as trim instead it just removes "--" from the beginning of the
--   string.
-- Comments: Only used in read_line
-- See Also: read_line
--------------------------------------------------------------------------------
global function trim_comment(sequence string)

    sequence tmp
    
    tmp = ""

    for i=1 to length(string) do
    if string[i] != '-' then
        tmp = string[i..length(string)]
        exit
    -- Else do next iteration
    end if
    end for

    return tmp

end function  -- trim_comment




--------------------------------------------------------------------------------
-- Routine: read_line
-- Syntax:  read_line(sequence command, sequence line)
-- Description: Reads the line and gets the data for documentating program.
-- See Also: generate_html
--------------------------------------------------------------------------------
global procedure read_line(sequence command, sequence line)

    sequence original_command  -- Holds the original command in case command 
                               -- isn't a valid data tag


    original_command = ""

    --Parsing = TRUE

    -- Check for empty command which means put data in last field accessed
    if length(command) = 0 then -- command = ""
    if Parse_type = LIBRARY then
        Library_header[Last] &= trim_comment(line)
    else -- Parse_type = ROUTINE
        if Last = SYNTAX then
                Routines[length(Routines)][SYNTAX] &=
                  (trim_comment(line) & "<br>")
        else
                Routines[length(Routines)][Last] &= trim_comment(line)
        end if
    end if

    return -- Stop here if comment is empty

    else
        original_command = command
        command = lower(trim(command))
    end if


    if equal(command, "name") then
    if Parse_type = ROUTINE then
        error(" \"Name:\" tag used after library name has" &
          " already been defined.")
    else
        Library_header[NAME] = line
        Last = NAME
        Parsing = TRUE
    end if
    
    elsif equal(command, "version") then
    if Parse_type = ROUTINE then
        error(" \"Version:\" tag used after library data has" &
          " already been defined.")
    else
        Library_header[VERSION] = line
        Last = VERSION
        Parsing = VERSION
    end if

    elsif equal(command, "license") then
    if Parse_type = ROUTINE then
        error("\"License:\" tag used after library data" & 
          " should have been defined.")
    else
        Library_header[LICENSE] = line
        Last = LICENSE
        Parsing = TRUE
    end if

    elsif equal(command, "intro") then
    if Parse_type = ROUTINE then
        error("\"Intro:\" tag used after library data" &
          " should have been defined.")
    else
        Library_header[INTRO] = line
        Last = INTRO
        Parsing = TRUE
    end if

    elsif equal(command, "routine") or equal(command, "constant")
      or equal(command, "type")then
    if Parse_type = LIBRARY then
        Parse_type = ROUTINE
    end if

    Routines = append(Routines, ROUTINE_STUB)
    Routines[length(Routines)][NAME] = line
    Last = ROUTINE
    Parsing = TRUE

    elsif equal(command, "author") then
    if Parse_type = LIBRARY then
        Library_header[AUTHOR] = line
    else
        Routines[length(Routines)][AUTHOR] = line
    end if
    Last = AUTHOR
    Parsing = TRUE

    elsif equal(command, "created") then
    if Parse_type = LIBRARY then
        Library_header[CREATED] = line
    else
        Routines[length(Routines)][CREATED] = line
    end if
    Last = CREATED
    Parsing = TRUE

    elsif equal(command, "updated") then
    if Parse_type = LIBRARY then
        Library_header[UPDATED] = line
    else
        Routines[length(Routines)][AUTHOR] = line
    end if
    Last = UPDATED
    Parsing = TRUE

    elsif equal(command, "platform") then
    if Parse_type = LIBRARY then
        error("\"Platform:\" tag can only be used to document routines.")
    else
        Routines[length(Routines)][PLATFORM] = line
        Last = PLATFORM
        Parsing = TRUE
    end if

    elsif equal(command, "syntax") then
    if Parse_type = LIBRARY then
        error("\"Syntax:\" tag can only be used to document routines.")
    else
        Routines[length(Routines)][SYNTAX] = (line & "<br>")
        Last = SYNTAX
        Parsing = TRUE
    end if


    elsif equal(command, "description") then
        if Parse_type = LIBRARY then
        error("\"Description:\" tag can only be used to document routines.")
    else
        Routines[length(Routines)][DESCRIPTION] = line
        Last = DESCRIPTION
        Parsing = TRUE
    end if

    elsif equal(command, "comments") then
    if Parse_type = LIBRARY then
        error("\"Comments:\" tag can only be used to document routines.")
    else
        Routines[length(Routines)][COMMENTS] = line
        Last = COMMENTS
        Parsing = TRUE
    end if

    elsif equal(command, "example") then
    if Parse_type = LIBRARY then
        error("\"Example:\" tag can only be used to document routines.")
    else
        Routines[length(Routines)][EXAMPLE] = line
        Last = EXAMPLE
        Parsing = TRUE
    end if

    elsif equal(command, "example program") then
    if Parse_type = LIBRARY then
        error("\"Example Program:\" tag can only be used to document " &
            " routines.")
    else
        Routines[length(Routines)][EXAMPLE_PROG] = line
        Last = EXAMPLE_PROG
        Parsing = TRUE
    end if

    elsif equal(command, "see also") then
    if Parse_type = LIBRARY then
        error("\"See Also:\" tag can only be used to document routines.")
    else
        Routines[length(Routines)][SEE_ALSO] = line
        Last = SEE_ALSO
        Parsing = TRUE
    end if

    else -- Not a valid data tag so treat as a normal comment
        if Parsing = TRUE then
            if Parse_type = LIBRARY then
                Library_header[Last] &= original_command & trim_comment(line)
            else -- Parse_type = ROUTINE
                Routines[length(Routines)][Last] &= original_command &
                  trim_comment(line)
            end if
        end if

    end if


end procedure  -- read_line




--------------------------------------------------------------------------------
-- Routine: comment
-- Syntax:  bool = comment(sequence line)
-- Description: Returns TRUE is line is a comment, FALSE otherwise
--------------------------------------------------------------------------------
global function comment(sequence line)

    -- Check to see if the line starts with a comment
    for j=1 to length(line) do

        if line[j] = '-' then
            if line[j+1] = '-'  then
                return j+2
            else
            return 0
            end if
        elsif line[j] != ' ' and line[j] != '\t' then
            return 0
        end if

    end for
    
    return 0

end function  -- comment




--------------------------------------------------------------------------------
-- Routine: see_also_links
-- Syntax:  s = see_also_links(sequence filename, sequence routine_names)
-- Description:
--   Takes the name of a file and a string of routine names and returns a 
--   string of HTML that has links to every routine in the file with that the
--   name passed. Needed for parsing "See Also" info.
-- See Also:  generate_html
--------------------------------------------------------------------------------
function see_also_links(sequence file, sequence string)

    sequence word, -- Holds each name while parsing
             text  -- The HTML text returned
    
    word = ""
    text = ""

    -- Parse string by commas ignoring spaces and "()"s
    for i=1 to length(string) do
        if string[i] = ',' then
            if length(word) > 0 then  -- word is not empty
                text &= sprintf("<a href=\"%s#%s\">%s</a>,\n",
                  {file, trim(word), trim(word)})
                word = ""
            end if
        elsif string[i] != ' ' and string[i] != '(' and string[i] != ')' then
           word &= string[i]
        end if
    end for

    -- Last word not separated by a ','
    if length(word) > 0 then
        text &= sprintf("<a href=\"%s#%s\">%s</a>\n",
          {file, trim(word), trim(word)})
    end if
    
    return text

end function  -- see_also_links





--------------------------------------------------------------------------------
-- Routine: generate_html
-- Syntax:  generate_html()
-- Description: 
--   Takes the data colected in Library_header and Routines and creates the
--   HTML documentation from it that looks like the documentation created by 
--   RDS.
--
-- Comments:  Very ugly coding when it comes to outputting the HTML
--------------------------------------------------------------------------------
global procedure generate_html()

    sequence output_file,   -- Name of the HTML documentation to be created
             list_file,     -- Name of the HTML file that holds the alphabetical
                            --  list of routines
             sorted_names,  -- List of routine names sorted alphabetically
             sorted_indeces -- List of the indeces of the data for each sorted
                            --  routine name

    integer fn1,   -- File handle for first file
            fn2,   -- File handle for second file
            index,  -- Temporary index used for generating the name of the 
                   --  sencond file
            ix                 
    object tmp     -- Temporary object, holds data for testing


    -- Open file to output to
    output_file = ""

    -- Loop until a suitable length name is returned
    while TRUE do
        output_file =
          prompt_string("File name to output to: ")

        if length(output_file) > 11 then
            puts(2, "File name is too long (8 + 3 chars max). Try again.\n")
        else
            exit
        end if

    end while

    -- Open file to output the alphabetical list to 
    -- (it's the first 4 characters of output_file with "_a_z.html" appended)

    index = match(".", output_file)
    
    if index > 0 then
        list_file = output_file[1..index]
    else
        list_file = output_file
    end if

    if length(list_file) <= 4 then
        list_file = list_file & "_a_z.htm"
    else
        list_file = list_file[1..4] & "_a_z.htm"
    end if


    -- Open files for output

    fn1 = open(output_file, "w")
    if fn1 = -1 then
        error(sprintf("File \"%s\" already exists or can not be created" &
          " for output.", {output_file}) )
    end if

    fn2 = open(list_file, "w")
    if fn2 = -1 then
        error(sprintf("File \"%s\" already exists or can not be created" &
          " for output.", {list_file}) )
    end if


    -- Title and intro

    printf(fn1, "<html>\n<head>\n<title>%s</title>\n", {Library_header[NAME]})
    -- Other stuff might go here (style, etc.)
    puts(fn1, "</head>\n" &
      "<body bgcolor=\"#FFFFFF\" link=\"#0033BB\" vlink=\"#006699\">\n\n" &
      "<center><font color=\"#006633\" size=\"+2\"><b>\n<br>\n")
    printf(fn1, "   %s\n </b></font></center>", {Library_header[NAME]})
    puts(fn1, "<br>\n<br>\n")
    puts(fn1, "<center><font color=\"#FF0099\" face=\"Arial, Helvetica\"" &
        " size=\"+1\">\n")
    printf(fn1, "(c) %s<br><br>\n</font></center>\n", {Library_header[AUTHOR]})

    puts(fn1, "<center><font color=\"#FF0099\" size=\"+2\">\n<br>\n" &
      "   1. Introduction \n </font></center>\n<br><br>\n")
    printf(fn1, "%s \n<br>\n<hr>\n<br>\n", {Library_header[INTRO]})

    -- Routine names for contents

    puts(fn1, "<center><font color=\"#FF0099\" size=\"+2\">\n" &
      "   2. Routines\n</font></center>\n")


    -- Sort routine names alphabetically and print

    sorted_names = {}
    sorted_indeces = repeat("", length(Routines))

    for i=1 to length(Routines) do
        sorted_names = append(sorted_names, trim(Routines[i][NAME]))
    end for

    sorted_names = sort(sorted_names)


    for i=1 to length(sorted_names) do
        puts(fn1, "&nbsp;&nbsp;")
        printf(fn1, "<a href=\"%s#%s\">\n <b>\n  %s\n </b>\n</a>\n<br>\n",
          {list_file, sorted_names[i], sorted_names[i]} )
    end for
    

    -- End of first page

    puts(fn1, "<p>&nbsp;<p>")
    puts(fn1, "<center><font color=\"#006633\"" &
      "face=\"Arial, Helvetica\" size=\"+1\">\n")
    printf(fn1, "... continue \n<a href=\"%s\">", {list_file})
    puts(fn1, "  3. Alphabetical Listing of All Routines, From A to Z\n</a>\n" &
      "</font></center>\n")
      
    puts(fn1, "\n</body>\n</html>")
    close(fn1)



    -- Output Routine contents on second page
    
    -- Get Routines indeces for each routine name

    sorted_indeces = repeat("", length(sorted_names))

    for i=1 to length(sorted_names) do
        for j=1 to length(Routines) do
            if equal(sorted_names[i], trim(Routines[j][NAME])) then
                sorted_indeces[i] = j
            end if
        end for
    end for

    -- Print out start of file

    puts(fn2, "<html>\n<head>\n<title>\n")
    puts(fn2, Library_header[NAME])
    puts(fn2, "</title>\n</head>\n<body>\n")
    puts(fn2, "<center>\n <font color=\"#FF0099\" size=\"+2\">\n <br>\n")
    puts(fn2, "   3. Alphabetical Listing of all Routines\n </font>\n</center>")


    -- Print out Routines data using indeces

--  Modification by Andy Drummond  9 Sept 2005
--  The routines in the "alphabetic listing" were not sorted ...
--  this was because the loop below simply stepped through the list but didn't
--  print the element given by the sorted list. Thus I have used ix to read the
--  sorted index and replaced [i] with [ix] throughout the rest of the loop.

    for i=1 to length(sorted_indeces) do
        ix = sorted_indeces[i]
        
         -- Print routine/type/constant name

         tmp = trim(Routines[ix][NAME])

--  Addition by Andy Drummond 14 Sept 2005
--  The See Also whatsits in th ehtml file do NOT cause a jump to the routine.
--  It looks like the routine name has a LF at the end which the See Also doesn't
--  So I remove it ..

	if tmp[length(tmp)] = 10 then
	   tmp = tmp[1..length(tmp)-1]
	end if
	
         printf(fn2, "<a name=\"%s\"></a><font color=\"#006633\" size=\"+2\">" &
           "<br><center><b>%s</b></center></font><p>\n\n",
           {tmp, tmp})
         -- Start table
         puts(fn2, "<table border=\"0\" cellspacing=\"2\"" & 
           "cellpadding=\"2\">\n\n")

         -- Print syntax

         tmp = Routines[ix][SYNTAX]
         if sequence(tmp) then
             if length(tmp) > 0 then
                 printf(fn2, "<tr><td valign=\"top\"><b>Syntax:</b> </td>\n" &
                    "<td><font face = \"Arial, Helvetica\">" &
                    "  %s\n </font>\n </td></tr>\n\n", {tmp})
             end if
         end if

         -- Print description
         tmp = Routines[ix][DESCRIPTION]
         if sequence(tmp) then
             if length(tmp) > 0 then
                 printf(fn2, "<tr><td valign=\"top\"><b>Description:</b></td>" &
                    "\n<td> %s\n </font>\n </td></tr>\n\n", {tmp})
             end if
         end if

         -- Print comments
         tmp = Routines[ix][COMMENTS]
         if sequence(tmp) then
             if length(tmp) > 0 then
                 printf(fn2, "<tr><td valign=\"top\"><b>Comments:</b></td>\n" &
                    "<td> %s\n </font>\n </td></tr>\n\n", {tmp})
             end if
         end if

         -- Print example
         tmp = Routines[ix][EXAMPLE]
         if sequence(tmp) then
             if length(tmp) > 0 then
                 printf(fn2, "<tr><td valign=\"top\"><b>Example:</b></td>\n" &
                    "<td><font face = \"Arial, Helvetica\">" &
                    "  <pre>%s</pre>\n </font>\n </td></tr>\n\n", {tmp})
             end if
         end if

         -- Print example program
         tmp = Routines[ix][EXAMPLE_PROG]
         if sequence(tmp) then
             if length(tmp) > 0 then
                 printf(fn2, "<tr><td valign=\"top\">\n<b>\n  Example<br>\n" &
                   "  Program:\n</b>\n</td>\n" &
                    "<td>\n <font color=\"#5500FF\">\n" &
                    "  %s\n </font>\n </td></tr>\n\n", {tmp})
             end if
         end if
         
         -- Print see also info
         tmp = Routines[ix][SEE_ALSO]
         if sequence(tmp) then
             if length(tmp) > 0 then
                 -- Change string to string with links
                 tmp = see_also_links(list_file, tmp)
                 printf(fn2, "<tr><td valign=\"top\"><b>See Also:</b></td>\n" &
                    "<td>\n  %s\n </font>\n </td></tr>\n\n", {tmp})
             end if
         end if
         
         -- End table
         puts(fn2, " </table>\n <br> <br>\n")

    end for

--  End of changes by Andy Drummond

    -- End of file

    puts(fn2, "</body></html>")
    close(fn2)


end procedure  -- generate_html




--------------------------------------------------------------------------------
-- Routine: print_data
-- Syntax:  print_data(x)
-- Description: 
--   This routine outputs the Euphoria data for Library_header and Routines. If
--   x is a sequence then the data is output to a file with that name, if x is
--   an integer representing an open file then the data is outputted to that
--   file.
--
-- Comments: This routine is only used while debugging.
--------------------------------------------------------------------------------
global procedure print_data(object file)

    integer fn
    
    if sequence(file) then
    fn = open(file, "w")
    if fn = -1 then
        error(sprintf("File \"%s\" already exists or can not be created" &
          " for output.", {file}) )
    end if
    else -- file is a filehandle
        fn = file
    end if

    -- Print Library_header

    puts(fn, "Library_header:\n\n")
    --print(fn, Library_header)
    puts(fn, "Name: \n" & Library_header[NAME] & "\n\n")
    puts(fn, "Author: \n" & Library_header[AUTHOR] & "\n\n")
    puts(fn, "Date: \n" & Library_header[CREATED] & "\n\n")
    puts(fn, "Updated: \n" & Library_header[UPDATED] & "\n\n")
    puts(fn, "License: \n" & Library_header[LICENSE] & "\n\n")
    --puts(1, "Requires: \n" & Library_header[REQUIRES] & "\n\n") -- TAG REMOVED
    puts(fn, "Intro: \n" & Library_header[INTRO] & "\n\n")
    
    puts(fn, "\n\n\n\n")
    
    -- Print Routines

    puts(fn, "Routines:\n\n")
    --print(fn, Routines)

    -- !! Needs to be finished!


end procedure  -- print_data

